/*****************************************************************************
/*
                                  WASD.h

Main header file for WASD VMS HTTP daemon.


VERSION HISTORY
---------------
31-OCT-1999  MGD  v6.1.0, CGIplus "agents"
16-OCT-1999  MGD  v6.0.3, DEC TCP/IP v5.n compatibility definitions
10-JAN-1999  MGD  v6.0.0, proxy serving
19-SEP-1998  MGD  v5.3.0, disperse to module headers :^)
14-APR-1998  MGD  v5.0.1, XSSI extensions
07-JAN-1998  MGD  v5.0.0, HTTPd version 5 (Secure Sockets Layer, using SSLeay)
05-OCT-1997  MGD  v4.5.0, file cache
06-SEP-1997  MGD  v4.4.0, multiple server host names and ports
08-JUN-1997  MGD  v4.2.0, CGIplus
01-FEB-1997  MGD  v4.0.0, HTTPd version 4
01-AUG-1996  MGD  v3.3.0, quite a few changes since 3.0.n
01-DEC-1995  MGD  v3.0.0, HTTPd version 3
20-DEC-1994  MGD  v2.0.0, multi-threaded version
20-JUN-1994  MGD  v1.0.0, single-threaded version
*/
/*****************************************************************************/

#ifndef WASD_H_LOADED
#define WASD_H_LOADED 1

#include <descrip.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>
#include <rmsdef.h>

/* Internet-related header files */
#include <socket.h>
#include <in.h>
#include <netdb.h>
#include <inet.h>

/* define required values from UCX$INETDEF.H (Multinet does not supply one) */
#define INET_PROTYP$C_STREAM 1
#define INETACP$C_TRANS 2
#define INETACP_FUNC$C_GETHOSTBYNAME 1
#define INETACP_FUNC$C_GETHOSTBYADDR 2
#define UCX$C_AF_INET 2
#define UCX$C_DSC_ALL 2
#define UCX$C_FULL_DUPLEX_CLOSE 8192
#define UCX$C_REUSEADDR 4
#define UCX$C_SOCK_NAME 4
#define UCX$C_SOCKOPT 1
#define UCX$C_TCP 6

/* Alpha NO MEMBER ALIGNMENT unless otherwise specified */
#ifdef __ALPHA
#   pragma nomember_alignment
#endif

/**********/
/* macros */
/**********/

#define boolean int
#define true 1
#define false 0

#define FI_LI WASD_MODULE, __LINE__

#define CONFIG_AUTH_FILE_NAME  "HTTPD$AUTH"
#define CONFIG_FILE_NAME       "HTTPD$CONFIG"
#define CONFIG_MAP_FILE_NAME   "HTTPD$MAP"
#define CONFIG_MSG_FILE_NAME   "HTTPD$MSG"

/* allows the HTTPd to be operated with SYSPRV permanently enabled */
#define OPERATE_WITH_SYSPRV 0

/* this value must also include connections in a "keep-alive" state */
#define DEFAULT_CONCURRENT_CONNECT_MAX    50

#define DEFAULT_TIMEOUT_INPUT_MINUTES       1
#define DEFAULT_TIMEOUT_OUTPUT_MINUTES     10
#define DEFAULT_TIMEOUT_NOPROGRESS_MINUTES  3
#define DEFAULT_TIMEOUT_KEEPALIVE_SECONDS   0

/* helps ensure a client doesn't get into an infinite loop */
#define KEEPALIVE_MAX 100

#define DEFAULT_DCL_CGIPLUSIN_SIZE    3072
#define DEFAULT_DCL_SYSCOMMAND_SIZE   4096
#define DEFAULT_DCL_SYSOUTPUT_SIZE    4096
#define DEFAULT_FILE_BUFFER_SIZE      1024
#define DEFAULT_NET_READ_BUFFER_SIZE  2048
#define DEFAULT_OUTPUT_BUFFER_SIZE    4096
#define MAX_REQUEST_HEADER            8192

/* control pre-expired responses */
#define PRE_EXPIRE_ADMIN            1
#define PRE_EXPIRE_DELETE_ON_CLOSE  1
#define PRE_EXPIRE_INDEX_OF         0
#define PRE_EXPIRE_MAPPING_RULES    0
#define PRE_EXPIRE_MENU             1
#define PRE_EXPIRE_PUT              1
#define PRE_EXPIRE_SSI              0
#define PRE_EXPIRE_UPD              1
#define PRE_EXPIRE_UPD_TREE_OF      1
#define PRE_EXPIRE_WATCH            0

/* versions of the HTTP protocol */
#define HTTP_VERSION_NOT_SET   0
#define HTTP_VERSION_0_9      09
#define HTTP_VERSION_1_0      10
#define HTTP_VERSION_1_1      11
#define HTTP_VERSION_UNKNOWN  -1

/* constants used to identify HTTP methods */
#define HTTP_METHOD_DELETE   0x01
#define HTTP_METHOD_GET      0x02
#define HTTP_METHOD_HEAD     0x04
#define HTTP_METHOD_POST     0x08
#define HTTP_METHOD_PUT      0x10
#define HTTP_METHOD_CONNECT  0x20

/* constants for controlling logging */
#define LOGGING_BEGIN  1
#define LOGGING_END    2
#define LOGGING_OPEN   3
#define LOGGING_CLOSE  4
#define LOGGING_FLUSH  5
#define LOGGING_ENTRY  6

/* constants used by HTTPd supervisor timer */
#define TIMER_INPUT             1
#define TIMER_OUTPUT            2
#define TIMER_KEEPALIVE         3

/* constants used by the description module */
#define DESCRIPTION_ALL         0x01
#define DESCRIPTION_TEXT_PLAIN  0x02
#define DESCRIPTION_TEXT_HTML   0x04
/* this is returned in the first character the file type contains none */
#define DESCRIPTION_IMPOSSIBLE  0xff

/* default directory listing layout */
#define DEFAULT_DIR_LAYOUT "I__L__R__S:b__D"

/* "http:", "https:" or "connect:" service and request */
#define SCHEME_HTTP     1
#define SCHEME_HTTPS    2
#define SCHEME_CONNECT  3

/* maximum Kbytes (1024) that can be in a body to be POSTed or PUTted */
#define DEFAULT_PUT_KBYTES_MAX 250
/* number of versions retained for files POSTed or PUTed */
#define DEFAULT_PUT_VERSION_LIMIT 3

/* header fields for HTTP/1.0 keep alive response */
#define DEFAULT_KEEPALIVE_HTTP_HEADER \
        "Connection: Keep-Alive\r\nKeep-Alive:\r\n"

/* if true, stops HTTPd server account with SYSPRV PUT(etc)ing - DANGEROUS */
#define PUT_DISALLOW_SYSPRV 0

#define CONFIG_HOME_PAGES_MAX      16
#define CONFIG_README_FILES_MAX    16
#define CONFIG_SCRIPT_RUNTIME_MAX  16

#define SCRIPT_NAME_SIZE  128
#define METHOD_NAME_SIZE   16

/* W:RE,G:RE,O:RWED,S:RWED */
#define PUT_DEFAULT_FILE_PROTECTION 0xaa00

/* number of times a request can internally redirect */
#define REQUEST_REDIRECTION_MAX 5

/*************/
/* durations */
/*************/

/* 1000 = 1mS, 100 = 100uS, 10 = 10Us, 1 = 1uS (for really fast systems :^) */
#define DURATION_UNITS_USECS 1000

/* durations are to be in units of 0.1 milliseconds */
#define DURATION_UNITS_USECS 100

#if DURATION_UNITS_USECS == 1
#  define DURATION_DECIMAL_PLACES 6
#elif DURATION_UNITS_USECS == 10
#  define DURATION_DECIMAL_PLACES 5
#elif DURATION_UNITS_USECS == 100
#  define DURATION_DECIMAL_PLACES 4
#else
   /* defaults to milliseconds */
#  undef DURATION_UNITS_USECS
#  define DURATION_UNITS_USECS 1000
#  define DURATION_DECIMAL_PLACES 3
#endif

#define USECS_IN_A_SECOND 1000000

/*********************/
/* functional macros */
/*********************/

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))

/* is it linear-white-space? (avoids the call overhead of an isspace()) */
#define ISLWS(c) (c == ' ' || c == '\t')

/* end of CR or LF terminated line? */
#define EOL(c) (c == '\r' || c == '\n' || !*cptr)
#define NOTEOL(c) (c != '\r' && c != '\n' && *cptr)

/*************************************************/
/* macros for development timing in microseconds */
/*************************************************/

/* 0..9 timers available */

#define USEC_TIMER_STORAGE \
   unsigned long  uSecSubx2 = 2, \
                  uSecEdiv10 = 10; \
   unsigned long  uSecResult, \
                  uSecDuration, \
                  uSecRemainder; \
   unsigned long  uSecBeginQuad [10][2], \
                  uSecEndQuad [10][2];

#define USEC_TIMER_STORAGE_EXTERN \
   extern unsigned long  uSecSubx2, \
                         uSecEdiv10; \
   extern unsigned long  uSecResult, \
                         uSecDuration, \
                         uSecRemainder; \
   extern unsigned long  uSecBeginQuad [10][2], \
                         uSecEndQuad [10][2];

#define USEC_TIMER_BEGIN(IDX) \
{ \
   sys$gettim (&uSecBeginQuad[IDX]); \
}

#define USEC_TIMER_END(IDX,WHERE) \
{ \
   sys$gettim (&uSecEndQuad[IDX]); \
   lib$subx (&uSecEndQuad[IDX], &uSecBeginQuad[IDX], &uSecResult, &uSecSubx2); \
   lib$ediv (&uSecEdiv10, &uSecResult, &uSecDuration, &uSecRemainder); \
   fprintf (stdout, "%s uS: %d\n", WHERE, uSecDuration); \
}

/*******************/
/* data structures */
/*******************/

struct VmsItem
{
   unsigned short  buf_len;
   unsigned short  item;
   unsigned char   *buf_addr;
   unsigned short  *ret_len;
};

struct AnExitHandler
{
   unsigned long  Flink;
   unsigned long  HandlerAddress;
   unsigned long  ArgCount;
   unsigned long  ExitStatusPtr;
};

struct AnIOsb
{
   unsigned short  Status;
   unsigned short  Count;
   unsigned long  Unused;
};

struct MbxSenseIOsb
{
   unsigned short  Status;
   unsigned short  MessageCount;
   unsigned long  MessageBytes;
};

/*
   Store these structures naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

/* per-request memory heap */
struct HeapStruct
{
   struct HeapStruct  *PrevPtr;
   struct HeapStruct  *NextPtr;
   unsigned int  ChunkSize;
   /* the actual data follows immediately after the length */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct ContentTypeStruct
{
   struct ContentTypeStruct  *NextPtr,
                             *HashCollisionPtr;
   boolean  ContentTypeUnknown;
   char  *SuffixPtr;
   char  *ContentTypePtr;
   char  *AutoScriptNamePtr;
   char  *ContentDescriptionPtr;
   char  *IconPtr;
   /* string data is stored here */
};

/*
This structure is used to both store and to point at the various components
depending on the context.
*/

struct IconStruct
{
   struct IconStruct  *NextPtr,
                      *HashCollisionPtr;
   /* string data is stored here */
};

struct  ListEntryStruct
{
   struct ListEntryStruct  *PrevPtr;
   struct ListEntryStruct  *NextPtr;
};

struct  ListHeadStruct
{
   struct ListEntryStruct  *HeadPtr;
   struct ListEntryStruct  *TailPtr;
   int  EntryCount;
};

/* these macros operate on the list head */
#define LIST_HEAD(lp) ((char*)((struct ListHeadStruct*)lp)->HeadPtr)
#define LIST_TAIL(lp) ((char*)((struct ListHeadStruct*)lp)->TailPtr)
#define LIST_COUNT(lp) (((struct ListHeadStruct*)lp)->EntryCount)
#define LIST_EMPTY(lp) (!((struct ListHeadStruct*)lp)->EntryCount)

/* these macros operate on list entries */
#define LIST_DATA(lp) ((char*)((char*)lp+sizeof(struct ListEntryStruct)))
#define LIST_PREV(lp) (((struct ListEntryStruct*)lp)->PrevPtr)
#define LIST_NEXT(lp) (((struct ListEntryStruct*)lp)->NextPtr)
#define LIST_HAS_PREV(lp) ((((struct ListEntryStruct*)lp)->PrevPtr) != NULL)
#define LIST_HAS_NEXT(lp) ((((struct ListEntryStruct*)lp)->NextPtr) != NULL)

/**********************************/
/* request history data structure */
/**********************************/

/* 256 bytes (a somewhat arbitrary quantity) */
struct  RequestHistoryStruct
{
   struct  ListEntryStruct  HistoryListEntry;
   struct  ServiceStruct  *ServicePtr;
   unsigned long  BinaryTime [2];
   unsigned long  BytesRawRx,
                  BytesRawTx,
                  ConnectNumber,
                  ResponseDuration;
   int  ResponseStatusCode;
   char  *RequestPtr;
   boolean  Truncated;
   char  ClientAndRequest [256-12-8-4-4-4-4-4-4-4];
};

/***************************/
/* configuration structure */
/***************************/

/* must be powers of two! */
#define  CONTENT_TYPE_HASH_TABLE_SIZE  128
#define  CONTENT_TYPE_HASH_TABLE_MASK  CONTENT_TYPE_HASH_TABLE_SIZE-1
#define  ICON_HASH_TABLE_SIZE  64
#define  ICON_HASH_TABLE_MASK  ICON_HASH_TABLE_SIZE-1

struct ConfigStruct
{
   boolean  AuthBasicEnabled,
            AuthDigestEnabled,
            CacheEnabled,
            DclFullRequest,
            DclSpawnAuthPriv,
            DirAccess,
            DirAccessSelective,
            DirNoImpliedWildcard,
            DirNoPrivIgnore,
            DirOwnerEnabled,
            DirPreExpired,
            DirReadMeBottom,
            DirReadMeTop,
            DirWildcardEnabled,
            DnsLookup,
            ErrorRecommend,
            ErrorSourceInfo,
            IncludeCommentedInfo,
            LoggingEnabled,
            LogPerService,
            MonitorEnabled,
            ProxyAddForwardedByEnabled,
            ProxyCacheEnabled,
            ProxyReportCacheLog,
            ProxyReportLog,
            ProxyServingEnabled,
            SsiEnabled,
            SsiAccessesEnabled,
            SsiExecEnabled;

   int  AcceptHostsLength,
        ActivityNumberOfDays,
        AuthCacheMinutes,
        AuthDigestNonceGetLifeTime,
        AuthDigestNoncePutLifeTime,
        AuthFailureLimit,
        AuthRevalidateUserMinutes,
        BufferSizeDclCommand,
        BufferSizeDclOutput,
        BufferSizeDclCgiPlusIn,
        BufferSizeNetRead,
        BufferSizeNetWrite,
        Busy,
        CacheEntriesMax,
        CacheChunkKBytes,
        CacheFileKBytesMax,
        CacheFrequentHits,
        CacheFrequentSeconds,
        CacheHashTableEntries,
        CacheTotalKBytesMax,
        CacheValidateSeconds,
        DclCgiPlusLifeTime,
        DclSubprocessHardLimit,
        DclSubprocessSoftLimit,
        DclZombieLifeTime,
        DECnetConnectListMax,
        DECnetReuseLifeTime,
        DirDescriptionLines,
        ErrorReportPathLength,
        HomePageCount,
        KeywordSearchScriptLength,
        LogExcludeHostsLength,
        MapUserNameCacheEntries,
        ProxyCacheDeviceMaxPercent,
        ProxyCacheDevicePurgePercent,
        ProxyCacheDeviceCheckMinutes,
        ProxyCacheFileKBytesMax,
        ProxyCacheRoutineHourOfDay,
        ProxyHostCachePurgeHours,
        ProxyHostLookupRetryCount,
        PutMaxKbytes,
        PutVersionLimit,
        ReadMeFileCount,
        RejectHostsLength,
        RequestHistory,
        ScriptRunTimeCount,
        ServerPort,
        ServiceLength,
        StreamLfConversionMaxKbytes,
        TimeoutMinutesInput,
        TimeoutMinutesNoProgress,
        TimeoutMinutesOutput,
        TimeoutSecondsKeepAlive;

   int  ScriptRunTimeLengthArray [CONFIG_SCRIPT_RUNTIME_MAX];

   struct ContentTypeStruct  
      *ContentTypeHashTable [CONTENT_TYPE_HASH_TABLE_SIZE];
   struct IconStruct *IconHashTable [ICON_HASH_TABLE_SIZE];

   char  CharsetDefault [32],
         DefaultDirLayout [32],
         ErrorReportPath [128],
         KeywordSearchScript [128],
         LogFileName [128],
         LogFormat [128],
         LogNaming [32],
         LogPeriod [32],
         ProxyCachePurgeList [64],
         ProxyCacheReloadList [64],
         ServiceNotFoundUrl [128];

   char  *AcceptHostsPtr,
         *LogExcludeHostsPtr,
         *RejectHostsPtr,
         *ServicePtr;

   char  *HomePageArray [CONFIG_HOME_PAGES_MAX],
         *ReadMeFileArray [CONFIG_README_FILES_MAX],
         *ScriptRunTimeArray [CONFIG_SCRIPT_RUNTIME_MAX];

   struct ContentTypeStruct  *ContentTypeListHeadPtr,
                             *ContentTypeListTailPtr;
   char  *ContentTypeDefaultPtr;
   int  ContentTypeStructOverhead;

   struct IconStruct  *IconListHeadPtr,
                      *IconListTailPtr;
   int  IconStructOverhead;

   /* only used while reading configuration from file */
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   char  *ProblemReportPtr;
   char  DirectiveName [32],
         LoadFileName [256];
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];

   /* only used when reporting or changing configuration */
   void  *RequestPtr;
};

/*************************/
/* per-server accounting */
/*************************/

/* as of v4.4 this structure is now > 255 bytes requiring two logicals */

struct AccountingStruct
{
   /* this is used to check for changes in HTTPd version requiring zeroing */
   unsigned long  AccountingVersion;
   /* this MUST be the first member (or so) to allow zeroing of the rest */
   unsigned long  ZeroedCount;

   unsigned long  AuthBasicCount,
                  AuthAgentCount,
                  AuthAuthorizedCount,
                  AuthDigestCount,
                  AuthHtDatabaseCount,
                  AuthSimpleListCount,
                  AuthNotAuthenticatedCount,
                  AuthNotAuthorizedCount,
                  AuthVmsCount,
                  CacheHitCount,
                  CacheHitNotModifiedCount,
                  CacheLoadCount,
                  ConnectAcceptedCount,
                  ConnectCount,
                  ConnectSslCount,
                  ConnectCurrent,
                  ConnectPeak,
                  ConnectRejectedCount,
                  ConnectTooBusyCount,
                  DclCgiPlusReusedCount,
                  DclSpawnCount,
                  DoAutoScriptCount,
                  DoCgiPlusScriptCount,
                  DoDclCommandCount,
                  DoDECnetCount,
                  DoDECnetReuseCount,
                  DoDECnetCgiCount,
                  DoDECnetCgiReuseCount,
                  DoDECnetOsuCount,
                  DoDECnetOsuReuseCount,
                  DoDirectoryCount,
                  DoIsMapCount,
                  DoScriptCount,
                  DoFileCount,
                  DoFileNotModifiedCount,
                  DoMenuCount,
                  DoPutCount,
                  DoServerAdminCount,
                  DoSsiCount,
                  DoUpdateCount,
                  LastExitStatus,
                  MethodDeleteCount,
                  MethodGetCount,
                  MethodHeadCount,
                  MethodPostCount,
                  MethodPutCount,
                  MethodTraceCount,
                  RedirectLocalCount,
                  RedirectRemoteCount,
                  RequestTotalCount,
                  RequestErrorCount,
                  RequestForbiddenCount,
                  RequestKeepAliveCount,
                  RequestKeepAliveMax,
                  RequestParseCount,
                  RequestPragmaNoCacheCount,
                  RequestProxyCount,
                  ResponseDurationCount,
                  ResponseDurationMax,
                  ResponseDurationMin,
                  StartupCount,
                  StreamLfConversionCount;

   unsigned long  QuadBytesRawRx [2],
                  QuadBytesRawTx [2],
                  QuadResponseDuration [2];

   unsigned long  ResponseStatusCodeCountArray [6];
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

/************************/
/* module include files */
/************************/

#ifndef ADMIN_H_LOADED
#   include "admin.h"
#endif
#ifndef AUTH_H_LOADED
#   include "auth.h"
#endif
#ifndef BASIC_H_LOADED
#   include "basic.h"
#endif
#ifndef CACHE_H_LOADED
#   include "cache.h"
#endif
#ifndef CGI_H_LOADED
#   include "cgi.h"
#endif
#ifndef CONFIG_H_LOADED
#   include "config.h"
#endif
#ifndef CONTROL_H_LOADED
#   include "control.h"
#endif
#ifndef DESCR_H_LOADED
#   include "descr.h"
#endif
#ifndef DCL_H_LOADED
#   include "dcl.h"
#endif
#ifndef DIR_H_LOADED
#   include "dir.h"
#endif
#ifndef DECNET_H_LOADED
#   include "decnet.h"
#endif
#ifndef ERROR_H_LOADED
#   include "error.h"
#endif
#ifndef FILE_H_LOADED
#   include "file.h"
#endif
#ifndef GRAPH_H_LOADED
#   include "graph.h"
#endif
#ifndef HTADMIN_H_LOADED
#   include "htadmin.h"
#endif
#ifndef HTTPD_H_LOADED
#   include "httpd.h"
#endif
#ifndef ISMAP_H_LOADED
#   include "ismap.h"
#endif
#ifndef MAPURL_H_LOADED
#   include "mapurl.h"
#endif
#ifndef MENU_H_LOADED
#   include "menu.h"
#endif
#ifndef MSG_H_LOADED
#   include "msg.h"
#endif
#ifndef NET_H_LOADED
#   include "net.h"
#endif
#ifndef PROXY_H_LOADED
#   include "proxy.h"
#endif
#ifndef PROXYCACHE_H_LOADED
#   include "proxycache.h"
#endif
#ifndef PROXYMAINT_H_LOADED
#   include "proxymaint.h"
#endif
#ifndef PUT_H_LOADED
#   include "put.h"
#endif
#ifndef REQUEST_H_LOADED
#   include "request.h"
#endif
#ifndef SSI_H_LOADED
#   include "ssi.h"
#endif
#ifndef SESOLA_H_LOADED
#   include "sesola.h"
#endif
#ifndef STMLF_H_LOADED
#   include "stmlf.h"
#endif
#ifndef SUPPORT_H_LOADED
#   include "support.h"
#endif
#ifndef UPD_H_LOADED
#   include "upd.h"
#endif
#ifndef VM_H_LOADED
#   include "vm.h"
#endif
#ifndef WATCH_H_LOADED
#   include "watch.h"
#endif

/*****************************/
/* per-thread data structure */
/*****************************/

/*
   Store these structure naturally-aligned on AXP.
   Uses a bit more storage but makes access as efficient as possible.
*/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

struct RequestStruct
{
   /***************************/
   /* general purpose storage */
   /***************************/

   /*
      Storage areas at the beginning of this structure, specifically
      above the field 'RetainAboveZeroBelow', are retained between
      successive requests during persistent connections.
      When a "keep-alive" connection is reinitialized ready for a
      potential new request all of this structure below (and including)
      that field are zeroed, effectively allowing a new request.
   */

   /* for maintaining the list of request structures */
   struct  ListEntryStruct  RequestListEntry;

   /* pointer to service structure of host/port etc of service connected to */
   struct ServiceStruct  *ServicePtr;

   int  /* counter for input timeout */
        TimerInputCount,
        /* counter for keep-alive timeout */
        TimerKeepAliveCount,
        /* previous Tx bytes for output no-progress checking */
        TimerNoProgressBytesTx,
        /* counter for output no-progress */
        TimerNoProgressCount,
        /* storage for server value (WATCH must be able to manipulate this) */
        TimerNoProgressSeconds,
        /* counter for output timeout */
        TimerOutputCount,
        /* supervisor timer has expired */
        TimerTerminatedCount;

   unsigned long  /* zone ID for request heap's virtual memory management */
                  VmHeapZoneId;

   unsigned short  /* channel to client's TCP/IP device */
                   ClientChannel,
                   /* client's port ;^) */
                   ClientPort;

   /* network read and write buffer */
   char  *NetReadRawDataPtr,
         *NetWriteRawDataPtr;
   int  NetReadRawDataSize,
        NetWriteRawDataLength;

   /* I/O status blocks */
   struct AnIOsb  NetReadIOsb,
                  NetWriteIOsb;

   /* network read and write AST function pointers */
   void (*NetReadRawAstFunctionPtr)(struct RequestStruct*);
   void (*NetWriteRawAstFunctionPtr)(struct RequestStruct*);

   unsigned long  /* client's binary IP address */
                  ClientIpAddress;

   char  /* internet host/domain name of client */
         ClientHostName [128],
         /* numeric host/domain address of client */
         ClientIpAddressString [32];

   int  /* server number of IP accept() and request creation */
        ConnectNumber,
        /* begin WATCHed */
        WatchItem,
        /* count of how many times the TCP/IP connection is reused */
        KeepAliveCount,
        /* an error is being reported from a redirect path */
        ErrorReportStatusCode,
        /* avoid redirection loops by keeping track of the number */
        RedirectionCount;

   char  /* when redirected this points to any authentication realm */
         *ErrorReportAuthRealmDescriptPtr;

   /* make it easy on ourselves, make the SSL structure pointer typeless! */
   void  *SeSoLaPtr;
   /* "http:" or "https:" (SSL) */
   int  RequestScheme;
   /* pointer to either "http:" or "https:" */
   char  *RequestSchemeNamePtr;

   /* the rest of this structure can be zeroed with persistent connections */
   /**********************************/
   unsigned long  RetainAboveZeroBelow;
   /**********************************/

   boolean  /* cache is currently being loaded for this path/file */
            CacheLoading,
            /* "temporary" file, delete it on close */
            DeleteOnClose,
            /* report any errors using the error report script */
            ErrorReportByRedirect,
            /* this request has a "Connection: Keep-Alive" token */
            KeepAliveRequest,
            /* the server has responded with "keep-alive" */
            KeepAliveResponse,
            /* this persistent connection has timed out */
            KeepAliveTimeout,
            /* escapes HTML-forbidden characters as sent to client */
            NetWriteBufferedEscapeHtml,
            /* response header has been sent to client */
            ResponseHeaderSent,
            /* response include an "Expires:" the same as "Last-Modified:" */
            ResponsePreExpired,
            /* "Pragma: no-cache" HTTP request header line encountered */
            PragmaNoCache,
            /* only authorize the script, not any path component */
            SetPathAuthorizeOnce,
            /* rule mapping, SET the path to be pre-expired */
            SetPathExpired,
            /* rule mapping, SET the path not-to-be-cached */
            SetPathNoCache,
            /* rule mapping, SET the path not-to-be-logged */
            SetPathNoLog,
            /* VMS user profile does not apply to this path */
            SetPathNoProfile,
            /* rule mapping, SET the path allowing stream-LF conversion */
            SetPathStmLF,
            /* an "Xray" returns full response as plain-text (diagnostic) */
            Xray;


   int  /* counter representing request activity has be incremented */
        AccountingDone,
        /* could not have come from cache, could have but didn't, did! */
        CacheInvolvement,
        /* multi-purpose, when loading and reading cache */
        CacheContentCount,
        /* count of bytes received so far in a POSTed or PUTed body */
        ContentCount,
        /* "Content-Length:" value */
        ContentLength,
        /* "Content-Type:" length */
        ContentTypeLength,
        /* size of buffer space for sys$output from DCL subprocess */
        *DclSysOutputSize,
        /* processing request header, count to find first blank line */
        HeaderConsecutiveNewLineCount,
        /* "Accept:" length */
        HttpAcceptLength,
        /* "Accept-Charset:" length */
        HttpAcceptCharsetLength,
        /* "Accept-Encoding:" length */
        HttpAcceptEncodingLength,
        /* "Accept-Language:" length */
        HttpAcceptLangLength,
        /* "Forwarded:" line(s) length */
        HttpForwardedLength,
        /* length of error report text/query-string */
        ErrorReportLength,
        /* value representation of request HTTP method */
        HttpMethod,
        /* value representation of HTTP protocol version number */
        HttpVersion,
        /* length of request header field */
        HttpAuthorizationLength,
        /* length of request header field */
        HttpRefererLength,
        /* length of request header field */
        HttpUserAgentLength,
        /* length of mapped path */
        MappedPathLength,
        /* interge representing prefered message language */
        MsgLanguage,
        /* size of network read buffer */
        NetReadBufferSize,
        /* length of path from HTTP header */
        PathInfoLength,
        /* length of query string from HTTP header */
        QueryStringLength,
        /* length of remote user name */
        RemoteUserLength,
        /* lenth of resource (path plus query string) from HTTp header */
        ResourceLength,
        /* number of characters up until the blank line */
        RequestHeaderLength,
        /* characters in the request line (e.g. "GET /path HTTP/1.0\n") */
        RequestLineLength,
        /* count of home page file names tried */
        RequestHomePageIndex,
        /* used during HEAD method to determine when header ends */
        ResponseHeaderNewlineCount,
        /* buffer for data length */
        ResponseHeaderDataLength,
        /* length of reponse header */
        ResponseHeaderLength,
        /* numeric equivalent of the "200", "302", etc., HTTP status code */
        ResponseStatusCode,
        /* VMS status of most recent task */
        TaskStatus;

   unsigned long  /* count of REQUEST-RELATED bytes received from the client */
                  BytesRx,
                  /* count of REQUEST-RELATED bytes sent to the client */
                  BytesTx,
                  /* count of TOTAL bytes received from the client */
                  BytesRawRx,
                  /* count of TOTAL bytes sent to the client */
                  BytesRawTx,
                  /* buffer for hash value used when caching */
                  CacheHashValue,
                  /* response duration */
                  ResponseDuration;

   unsigned long  /* 64 bit VMS internal time, start of request processing */
                  BinaryTime [2],
                  /* 64 bit VMS time from "If-Modified-Since:" header */
                  IfModifiedSinceBinaryTime [2],
                  /* length paramater from "If-Modified-Since:" header */
                  IfModifiedSinceLength;

   unsigned short  /* component equivalent of 'BinaryTime' */
                   NumericTime [7];

   char  /* multi-purpose, when loading and reading cache */
         *CacheContentPtr,
         /* points into cache buffer when loading via record IO */
         *CacheContentCurrentPtr,
         /* points to dynamic buffer for body POSTed, PUT, etc. */
         *ContentBufferPtr,
         /* used as pointer into above buffer */
         *ContentPtr,
         /* buffer space for sys$output from DCL subprocess */
         *DclSysOutputPtr,
         /* pointer to other info when generating error report */
         *ErrorOtherTextPtr,
         /* pointer to error text/query-string */
         *ErrorReportPtr,
         /* pointer to explanation when generating error report */
         *ErrorTextPtr,
         /* "Accept:" header lines */
         *HttpAcceptPtr,
         /* "Accept-Charset:" header lines */
         *HttpAcceptCharsetPtr,
         /* "Accept-Encoding:" header lines */
         *HttpAcceptEncodingPtr,
         /* "Accept-Language:" header lines */
         *HttpAcceptLangPtr,
         /* pointer to heap storage of "Authorization:" field */
         *HttpAuthorizationPtr,
         /* stream's MIME content type */
         *HttpContentTypePtr,
         /* "Cookie:" request header encountered */
         *HttpCookiePtr,
         /* "ETag:" header line (only for proxy use) */
         *HttpETagPtr,
         /* "Forwarded:" header line */
         *HttpForwardedPtr,
         /* "Host:" header line */
         *HttpHostPtr,
         /* pointer to heap storage of "If-Modified-Since:" field */
         *HttpIfModifiedSincePtr,
         /* "Pragma:" request header encountered */
         *HttpPragmaPtr,
         /* "Proxy-Connection:" request header encountered */
         *HttpProxyConnectionPtr,
         /* pointer to heap storage of "Referer:" field */
         *HttpRefererPtr,
         /* pointer to heap storage of "User-Agent:" field */
         *HttpUserAgentPtr,
         /* pointer to heap storage of redirection location path */
         *LocationPtr,
         /* pointer to mapped path */
         *MappedPathPtr,
         /* pointer to heap storage client-read network buffering */
         *NetReadBufferPtr,
         /* pointer to heap storage of request path */
         *PathInfoPtr,
         /* pointer to heap storage of request query string */
         *QueryStringPtr,
         /* pointer to requested resource (path and query string) */
         *ResourcePtr,
         /* pointer to request header */
         *RequestHeaderPtr,
         /* pointer to response header */
         *ResponseHeaderPtr,
         /* buffer for data pointer */
         *ResponseHeaderDataPtr,
         /* rule mapping SET the path with a CGI variable prefix */
         *SetPathCgiPrefixPtr,
         /* rule mapping SET the path with a specific charset */
         *SetPathCharsetPtr,
         /* rule mapping SET the path with a specific content-type */
         *SetPathContentTypePtr,
         /* rule mapping SET the path with a "Index of" format */
         *SetPathIndexPtr;

   char  /* GMT-adjusted string equivalent of 'BinaryTime' */
         GmDateTime [48],
         /* string reqpresenting the request's HTTP method */
         HttpMethodName [METHOD_NAME_SIZE],
         /* authenticated user name from "Authorization:" */
         RemoteUser [AUTH_MAX_USERNAME_LENGTH+1],
         /* storage when authenticating remote user */
         RemoteUserPassword [AUTH_MAX_PASSWORD_LENGTH+1],
         /* parsed path storage */
         RequestFileName [256],
         /* */
         RequestMappedFile [256],
         /* */
         RequestMappedScript [256],
         /* mapped script name */
         ScriptName [SCRIPT_NAME_SIZE];

   /* request is being used to fill a cache entry */
   struct CacheStruct  *CachePtr;

   /* AST function pointers */
   void (*ResponseHeaderAstFunctionPtr)(struct RequestStruct*);

   /* */
   struct ContentTypeStruct  ContentType;

   /* RMS structures used for various file processing purposes */
   struct FAB  RequestParseFab;
   struct NAM  RequestParseNam;

   /*****************/
   /* scripting/DCL */
   /*****************/

   boolean  /* ensure each record has a trailing newline */
            CgiAddNewline,
            /* */
            CgiBufferRecords,
            /* check the first record of the response header */
            CgiCheckedHeader,
            /* */
            CgiContentTypeText,
            /* still in the response header */
            CgiInHeader,
            /* script output escape is being processed */
            CgiPlusCalloutInProgress,
            /* true if a mapping rule identifies this as a "CGIplus" script */
            IsCgiPlusScript;

   int  /* length of heap memory pointed to by CgiBufferPtr */
        CgiBufferLength,
        /* remaining space in heap memory pointed to by CgiBufferPtr */
        CgiBufferRemaining,
        /* count of characters pointed to by 'CgiPlusCalloutOutputPtr' */
        CgiPlusCalloutOutputCount,
        /* how many attempts have been made to restart a CGI script */
        CgiScriptRetryCount;

   char  /* heap memory in which the CGI variables are generated */
         *CgiBufferPtr,
         /* current position in CGI variable buffer */
         *CgiBufferCurrentPtr,
         /* pointer to a record output by a DCL script agent */
         *CgiPlusCalloutOutputPtr;

   /* AST function pointers */
   void (*WatchShowProcessNextTaskFunction)(struct RequestStruct*);

   /**************************/
   /* sentinal I/O sequences */
   /**************************/

   int  /* length of scripting end-of-file (output) sequence */
        SequenceEofLength,
        /* length of agent end-of-text sequence */
        SequenceEotLength,
        /* length of CGIplus scripting "escape" sequence */
        SequenceEscLength;

   char  /* scripting end-of-file (output) sequence */
         *SequenceEofPtr,
         /* agent end-of-text sequence */
         *SequenceEotPtr,
         /* CGIplus scripting "escape" sequence */
         *SequenceEscPtr;

   /**************/
   /* task lists */
   /**************/

   /* these two modules can have multiple, concurrent instances */
   struct  ListHeadStruct  DirTaskList,
                           SsiTaskList;

   /* these modules/tasks can only have one total or multiple serially */
   struct  DclTaskStruct      *DclTaskPtr;
   struct  DECnetTaskStruct   *DECnetTaskPtr;
   struct  DescrTaskStruct    *DescrTaskPtr;
   struct  DirTaskStruct      *DirTaskPtr;
   struct  FileTaskStruct     *FileTaskPtr;
   struct  GraphTaskStruct    *GraphTaskPtr;
   struct  HTAdminTaskStruct  *HTAdminTaskPtr;
   struct  IsMapTaskStruct    *IsMapTaskPtr;
   struct  MenuTaskStruct     *MenuTaskPtr;
   struct  ProxyTaskStruct    *ProxyTaskPtr;
   struct  PutTaskStruct      *PutTaskPtr;
   struct  SsiTaskStruct      *SsiTaskPtr;
   struct  UpdTaskStruct      *UpdTaskPtr;

   /****************************************/
   /* storage specific to output buffering */
   /****************************************/

   /* the head of the output buffer list */
   struct ListHeadStruct  OutputBufferList;

   /* pointer to heap storage structure for output buffering */
   struct ListEntryStruct  *OutputBufferStructPtr,
                           *OutputBufferNowStructPtr;

   /* pointer to any AST function */
   void  *OutputBufferAstFunctionPtr;

   boolean  /* all output buffers must be written to network */
            OutputBufferFlush;

   int  /* count of all data in all output buffers */
        OutputBufferCount,
        /* bytes still available in the current output buffer */
        OutputBufferRemaining;

   char  /* pointer to the start of the current buffer */
         *OutputBufferPtr,
         /* pointer used to point into 'OutputBuffer' storage */
         *OutputBufferCurrentPtr;

   /****************************************************/
   /* storage specific to authentication/authorization */
   /****************************************************/

   boolean  /* authenticated user may only access via https: (SSL) */
            AuthHttpsOnly,
            /* do not cache authentication information, revalidate each time */
            AuthNoCache,
            /* the authentication realm name is a synonym for the SYSUAF */
            AuthRealmIsVmsSynonym,
            /* the user was authenticated from the SYSUAF */
            AuthSysUafAuthenticated,
            /* user may change SYSUAF password */
            AuthSysUafCanChangePwd,
            /* user can authenticate even if hours seem to prohibit */
            AuthSysUafNilAccess;

   int  /* VMS status result of authentication/authorization */
        AuthorizationStatus,
        /* only when script is authentication agent */
        AuthAgentParameterLength,
        /* length of string */
        AuthGroupReadLength,
        /* length of string */
        AuthGroupWriteLength,
        /* length of string */
        AuthRealmLength,
        /* authenticated user details length */
        AuthUserDetailsLength,
        /* number of identifiers loaded into array */
        AuthVmsIdentifiersCount;

   unsigned short  /* length of the user profile pointed to */
                   AuthVmsUserProfileLength;

   unsigned long  /* value representation of scheme being used for challenge */
                  AuthChallengeScheme,
                  /* the capability flags of the authorization path */
                  AuthGroupCan,
                  /* VMS identifier value controlling read access */
                  AuthGroupReadVmsIdentifier,
                  /* VMS identifier value controlling write access */
                  AuthGroupWriteVmsIdentifier,
                  /* VMS identifier value controlling authentication */
                  AuthRealmVmsIdentifier,
                  /* the capability flags of the request (ACT ON THESE!) */
                  AuthRequestCan,
                  /* value representation of scheme being used for request */
                  AuthScheme,
                  /* source of read-only group information */
                  AuthSourceGroupRead,
                  /* source of full-access group information */
                  AuthSourceGroupWrite,
                  /* source of realm information */
                  AuthSourceRealm,
                  /* the capability flags of the remote user */
                  AuthUserCan,
                  /* what everybody else can do on that path */
                  AuthWorldCan;

   unsigned long  /* pointer to array of VMS identifiers (zero-terminated) */
                  *AuthVmsIdentifiersPtr;

   char  /* only when script is authentication agent */
         *AuthAgentParameterPtr,
         /* BASIC challenge response header field */
         *AuthBasicChallengePtr,
         /* DIGEST challenge response header field */
         *AuthDigestChallengePtr,
         /* when digest authenticating, the client-supplied nonce */
         *AuthDigestNoncePtr,
         /* when digest authenticating, the client-supplied digest */
         *AuthDigestResponsePtr,
         /* when digest authenticating, the client-supplied URI */
         *AuthDigestUriPtr,
         /* pointer to authorization group name */
         *AuthGroupReadPtr,
         /* pointer to string of IP-address/username masks */
         *AuthGroupRestrictListPtr,
         /* pointer to authorization group name */
         *AuthGroupWritePtr,
         /* pointer to authorization realm name */
         *AuthRealmPtr,
         /* pointer to authorization realm synonym string */
         *AuthRealmDescriptionPtr,
         /* authenticated user details */
         *AuthUserDetailsPtr,
         /* user profile is used by sys$check_access() */
         *AuthVmsUserProfilePtr,
         /* pointer to string of IP-address/username masks */
         *AuthWorldRestrictListPtr;

   char  /* basic, digest, etc. */
         AuthType [16];

   /* a pointer to the actual record in the cache itself */
   struct AuthCacheRecordStruct  *AuthCacheRecordPtr;
   /* a pointer to a template used when searching the cache */
   struct AuthCacheRecordStruct  *AuthCacheSearchRecordPtr;

   /* AST function pointers */
   void (*AuthorizeAstFunctionBuffer)(struct RequestStruct*);
   void (*AuthorizeAstFunction)(struct RequestStruct*);
   void (*AuthAgentCalloutFunction)(struct RequestStruct*);

   /***************/
   /* zeroed end! */
   /***************/

   unsigned long  ZeroedEnd;
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

#endif /* WASD_H_LOADED */

/*****************************************************************************/
